# Copyright 2014, 2015, 2016, 2017, 2018, 2019, 2020 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  See the file COPYING.  If not, see
# <http://www.gnu.org/licenses/>.


# This is various functions related to terdragon calculations, used for
# development and testing.  Everything likely to change, or move, etc.
#
# Most functions are named like in terdragon.tex.
#
# Coordinate functions like mul() and bk() act on integer $x,$y in the style
# of Math::PlanePath::TerdragonCurve, so representing a point $x/2 + $y*sqrt(3).

package MyTerdragon;
use 5.010; # for // operator
use strict;
use warnings;
use Carp 'croak';
use Math::BaseCnv 'cnv';
use Memoize;

use base 'Exporter';
our @EXPORT_OK = ('chamfer','transform_to_horizontal',
                  'to_bigint','bigint_class',
                  'R','V','A','AR','D','P',
                  'hull_expand','hull_and_end','hull_points_sort_by_angle',
                  'hull_expand_RV',
                  'P1','P2','P3','P4',
                  'P5','P6','P7','p',
                  'mul','div','xy_norm','bk',
                  'points_str','xy_to_node_name',
                  'is_integer','is_pow2',
                  'count_low_zeros_of_aref','count_low_twos_of_aref',
                  'path_tree_depth_to_hull_area',
                  'path_tree_depth_to_row_hull_area',
                  'xy_rotate_plus30','xy_rotate_minus30',
                  'xy_rotate_plus60','xy_rotate_minus60',
                  'xy_rotate_plus120','xy_rotate_minus120',
                  'xy_rotate_plus90','xy_rotate_minus90', 'xy_rotate_180',

                  '@dir3_to_dx','@dir3_to_dy',
                  '@dir6_to_dx','@dir6_to_dy',
                  'dxdy_to_dir3','dxdy_to_dir3_maybe',
                  'dxdy_to_dir6','triangle_normalize',
                  'BitAboveLowestOne',
                  'terdragon_Graph','triangle_turn_tree_Graph',
                  'triangle_area_and_connections_tree_Graph',
                  'triangle_area_Graph',

                  'turn',
                  'TurnLpred','TurnRpred','TurnLpred_flat','TurnRpred_flat',
                  'dNextL','dNextR','dNextL_flat','dNextR_flat',
                  'dNextLsub1_flat',

                  '@dNextL_state_table','@dNextR_state_table',
                  '@dNextLsub1_state_table','@dNextLsub1_state_to_name',
                  '@dNextLsub1_state_to_d','$dNextLsub1_start_state',

                  'dir',

                  'Rpred','Rpred_of_ternary_str',
                  'Rpred_flat','NonRpred_flat','@Rpred_state_table',
                  'Lpred_k','Lpred_k_of_ternary_str',
                  'Lpred_k_flat','NonLpred_k_flat','@Lpred_state_table',
                  'Bpred_k_flat','NonBpred_k_flat',

                  'RpointPred',
                  '$RpointPred_segs',
                  'RpointPred_init','$NonRpointPred_flat','$RpointPred_flat',
                  '@RpointPred_state_to_conf',

                  'R1sides_flat','L1sides_k_flat',
                  'R2sides_flat','L2sides_k_flat',
                  'Spred_k_flat','SpredRight_k_flat','SpredLeft_flat',
                  'Dpred_k_flat','DpredRight_k_flat','DpredLeft_flat',
                  'Tpred_k_flat','TpredRight_k_flat','TpredLeft_flat',
                  'Tperm',

                  '@EpredR_state_table','@EpredL_state_table',
                  'RsideNum1_flat','LsideNum1_flat',
                  'RsideNum2_flat','LsideNum2_flat',
                  'EpredR_flat','EpredL_flat',
                  'EpredL','EpredL_of_ternary_str',
                  'Epred_flat',
                  'EpredRpair_flat','EpredLpair_flat',
                  'EpredTwelve_LtoH_flat',

                  '$R_triangle_segs','$L_triangle_segs','$B_triangles_segs',

                  '$hullstouch_segs',
                  'fBpred_hullstouch_init', # the following variables
                  '@hullstouch_state_to_conf','@hullstouch_state_table',
                  '$fNonRpred_hullstouch_flat','$fNonLpred_hullstouch_flat',
                  '$fNonBpred_hullstouch_flat',
                  '$fRpred_hullstouch_flat','$fLpred_hullstouch_flat',
                  '$fBpred_hullstouch_flat',

                  '$hullsaround_segs_outer','$hullsaround_segs_inner',
                  '$hullsaround_segs_all',

                  'fBpred_hullsaround_init', # the following variables
                  '@hullsaround_state_to_conf','@hullsaround_state_table',
                  '$fNonRpred_hullsaround_flat','$fNonLpred_hullsaround_flat',
                  '$fNonBpred_hullsaround_flat',
                  '$fRpred_hullsaround_flat','$fLpred_hullsaround_flat',
                  '$fBpred_hullsaround_flat',

                  'fRpred_flat','fNonRpred_flat',
                  'fLpred_flat','fNonLpred_flat',
                  'fBpred_flat','fNonBpred_flat',

                  'AltEnd','alternate_hull_and_end',
                  'AltRpred_k','AltLpred_k',
                  'AltRpred3_flat',
                  'AltNonRpred9_flat_even','AltRpred9_flat_even',
                  'AltNonRpred3_flat_even','AltNonRpred3_flat_odd',

                  '$alt_hullstouch_segs','$alt_hullstouch_odd_segs',
                  'fAltBpred_hullstouch_init', # the following variables
                  '@alt_hullstouch_state_to_conf','@alt_hullstouch_state_table',
                  '$fAltNonRpred_hullstouch_flat',
                  '$fAltNonLpred_hullstouch_flat',
                  '$fAltNonBpred_hullstouch_flat',
                  '$fAltRpred_hullstouch_flat',
                  '$fAltLpred_hullstouch_flat',
                  '$fAltBpred_hullstouch_flat',

                  '@fAltBpred_state_table',
                  '@fAltBpred_state_names',
                  'fAltRpred_flat',
                 );
our %EXPORT_TAGS = ( all => \@EXPORT_OK );

# uncomment this to run the ### lines
# use Smart::Comments;



sub xy_norm {
  my ($x,$y) = @_;
  return ($x*$x + 3*$y*$y)/4;
}

use constant::defer TerdragonCurve_path => sub {
  require Math::PlanePath::TerdragonCurve;
  return  Math::PlanePath::TerdragonCurve->new;
};

#------------------------------------------------------------------------------
sub chamfer {
  my ($delta, $name) = @_;
  $name //= '\MyChamfer';
  if ($delta == 0) { return ''; }
  if ($delta == 1) { return "+$name"; }
  if ($delta == -1) { return "-$name"; }
  return sprintf '%+d*%s', $delta, $name;
}

sub transform_to_horizontal {
  my ($x,$y, $level) = @_;
  $level %= 12;
  while ($level >= 2) {
    ($x,$y) = xy_rotate_minus60($x,$y);
    $level -= 2;
  }
  if ($level) {
    ($x,$y) = mul($x,$y);
    ($x,$y) = xy_rotate_minus60($x,$y);
  }
  return ($x,$y);
}

sub to_bigint {
  my ($n) = @_;
  return bigint_class()->new($n);
}
use constant::defer bigint_class => sub {
  require Math::BigInt;
  eval { Math::BigInt->import (try => 'GMP') };
  return "Math::BigInt";
};

#------------------------------------------------------------------------------
# sequences

sub A {
  my ($k) = @_;
  return 2*AR($k);
}
sub AR {
  my ($k) = @_;
  return ($k == 0 ? 0 : 3**($k-1) - 2**($k-1));
}
sub P {
  my ($k) = @_;
  return ($k == 0 ? 2 : 3**($k-1) + 2**$k + 1);
}

# return the number of double-visited points in level $k
sub D {
  my ($k) = @_;
  return ($k == 0 ? 0 : 2**$k - 2);
}

# return the number of right boundary segments in curve $k
# left boundary segments are the same
sub R {
  my ($k) = @_;
  return ($k == 0 ? 1 : 3*2**($k-1));
}

# return the number of boundary segments in a V shape level $k
sub V {
  my ($k) = @_;
  return ($k == 0 ? 2 : 3*2**($k-1));
}


#------------------------------------------------------------------------------
# Math::Geometry::Planar things

# $points is an arrayref of [$x,$y] pairs.
# Return a string "x,y x,y ...".
sub points_str {
  my ($points) = @_;
  return join(' ',map{join(',',@$_)}@$points);
}

sub xy_to_node_name {
  my ($x,$y) = @_;
  if ($x==int($x) && $y==int($y)) {
    if ($x < 0) { $x = '-n'.-$x; }
    if ($y < 0) { $y = 'n'.-$y; }
    return "p$x-$y";
  } else {
    return "$x,$y";
  }
}


#------------------------------------------------------------------------------
sub is_integer {
  my ($n) = @_;
  return $n == int($n);
}
sub is_square {
  my ($n) = @_;
  my $s = int(sqrt($n));
  my $ss = $s*$s;
  if ($ss > $n) {
    $s--;
    $ss = $s*$s;
  }
  return $ss == $n;
}

sub is_pow2 {
  my ($n) = @_;
  (($n >> 1) < $n) or die "is_pow2() infinite";

  while ($n > 1) {
    if ($n & 1) {
      return 0;
    }
    $n >>= 1;
  }
  return ($n == 1);
}

sub count_low_zeros_of_aref {
  my ($digits) = @_;
  foreach my $i (0 .. $#$digits) {
    if ($digits->[$i] != 0) {
      return $i;
    }
  }
  return scalar(@$digits);
}
sub count_low_twos_of_aref {
  my ($digits) = @_;
  foreach my $i (0 .. $#$digits) {
    if ($digits->[$i] != 2) {
      return $i;
    }
  }
  return scalar(@$digits);
}


#------------------------------------------------------------------------------
# w^2 = -1+w
# c = (X-Y)/2  x=2c+d
# d = Y        y=d
# ((X-Y)/2 + Y*w)*(w+1)
# = (X-Y)/2 + ((X-Y)/2 + Y)*w + Y*w^2
# = (X-Y)/2 + ((X-Y)/2 + Y)*w + Y*(-1+w)
# = (X-Y)/2 - Y + ((X-Y)/2 + Y + Y)*w
# = (X-3*Y)/2 + (X+3*Y)/2 * w
# xm = 2*(X-3*Y)/2 + (X+3*Y)/2
#    = (2X - 6Y + X + 3Y)/2
#    = (3X - 3Y)/2
#
# GP-DEFINE  sqrt3i = quadgen(-12);
# GP-Test  sqrt3i^2 == -3
# GP-DEFINE  w6 = 1/2 + 1/2*sqrt3i;
# GP-DEFINE  b = w6 + 1;
# GP-Test  ('x + sqrt3i*'y)*b       == 3*('x-'y)/2 + ('x+3*'y)/2*sqrt3i
# GP-Test  ('x + sqrt3i*'y)*conj(b) == 3*('x+'y)/2 + (3*'y-'x)/2*sqrt3i
sub mul {
  my ($x,$y) = @_;
  return (3*($x-$y)/2, ($x+3*$y)/2);
}
sub mul_conj {
  my ($x,$y) = @_;
  return (3*($x+$y)/2, (3*$y-$x)/2);
}

sub div {
  my ($x,$y) = @_;
  ($x+$y) % 2 == 0 or die;
  (-$x+3*$y) % 6 == 0 or die;
  return (($x+$y)/2, (-$x+3*$y)/6);
}

our @dir3_to_dx = (2,-1,-1);
our @dir3_to_dy = (0, 1,-1);
our @dir6_to_dx = (2, 1,-1,-2, -1, 1);
our @dir6_to_dy = (0, 1, 1, 0, -1,-1);

# return $x,$y unfolded around pivot $ex,$ey
sub unfold_middle {
  my ($x,$y, $ex,$ey) = @_;
  ($x,$y) = xy_rotate_plus120($x,$y);
  return ($ex+$x, $ey+$y);
}
sub unfold_end {
  my ($x,$y, $ex,$ey) = @_;
  my ($mx,$my) = xy_rotate_plus60($ex,$ey);
  return ($mx+$x, $my+$y);
}

sub points_midpoint {
  my ($p1,$p2, $frac) = @_;
  $frac //= .5;
  return [ $p1->[0]*$frac + $p2->[0]*(1-$frac),
           $p1->[1]*$frac + $p2->[1]*(1-$frac) ];
}

sub hull_points_sort_by_angle {
  my ($points, $ex,$ey) = @_;
  return [ sort {hull_point_cmp_by_angle($a,$b, $ex,$ey)} @$points ];
}
# ascending relative to direction of $sx,$sy
sub hull_point_cmp_by_angle {
  my ($a,$b, $sx,$sy) = @_;
  require Math::NumSeq::PlanePathDelta;
  my $a4 = Math::NumSeq::PlanePathDelta::_dxdy_to_dir4(@$a);
  my $b4 = Math::NumSeq::PlanePathDelta::_dxdy_to_dir4(@$b);
  if ($a->[0] == 0 && $a->[1] == 0) { $a4 = 3.5; }
  if ($b->[0] == 0 && $b->[1] == 0) { $b4 = 3.5; }
  my $o4 = Math::NumSeq::PlanePathDelta::_dxdy_to_dir4($sx,$sy);
  # +.5 to allow for early PT1 around a bit
  $a4 = ($a4 - $o4 + .5);
  $b4 = ($b4 - $o4 + .5);
  while ($a4 < 0) { $a4 += 4; }
  while ($b4 < 0) { $b4 += 4; }
  return ($a4 <=> $b4);
}

# each x,y is a corner of a triangle
# return them in a normalized order
sub triangle_normalize {
  my ($x1,$y1, $x2,$y2, $x3,$y3) = @_;
  if ($y2 == $y3) {
    ($x1,$y1, $x2,$y2, $x3,$y3) = ($x2,$y2, $x3,$y3, $x1,$y1);
  }
  if ($y1 == $y3) {
    ($x1,$y1, $x2,$y2, $x3,$y3) = ($x1,$y1, $x3,$y3, $x2,$y2);
  }
  if ($x1 > $x2) {
    ($x1,$y1, $x2,$y2) = ($x2,$y2, $x1,$y1);
  }
  return ($x1,$y1, $x2,$y2, $x3,$y3);
}

#------------------------------------------------------------------------------
# Convex Hull

sub hull_expand {
  my ($points, $ex,$ey, %options) = @_;
  ### hull_expand(): "end=$ex,$ey  " . points_str($points)

  # if (@$points < 3) {
  #   ### small ...
  #   $points = [ @$points,
  #               # points_midpoint($points->[0],$points->[1],1/4),
  #               points_midpoint($points->[0],$points->[1],1/2),
  #               # points_midpoint($points->[0],$points->[1],3/4),
  #               ];
  # }

  # print " % at pxy=$ex,$ey  ",points_str($points),"\n";
  $points = [ @$points,
              (map{[ unfold_middle(@$_, $ex,$ey) ]} @$points),
              (map{[ unfold_end   (@$_, $ex,$ey) ]} @$points),
            ];

  require Math::Geometry::Planar;
  my $planar = Math::Geometry::Planar->new;
  $planar->points($points);
  if (@$points >= 5) {
    ### convexhull2 of: points_str($points)
    $planar = $planar->convexhull2;
    $points = $planar->points;
  }
  unless ($options{'unsorted'}) {
    $points = hull_points_sort_by_angle($points,
                                        xy_rotate_180($ex,$ey));
  }
  ($ex,$ey) = mul($ex,$ey);
  return ($points, $ex,$ey);
}

sub hull_and_end {
  my ($k) = @_;
  my $ex = 2;
  my $ey = 0;
  my $points = [ [0,0], [$ex,$ey] ];
  foreach (1 .. $k) {
    ($points,$ex,$ey) = hull_expand($points,$ex,$ey);
  }
  return ($points,$ex,$ey);
}

{
  my @p = ([-18,0], [-3, 15], [-3, -3], [0, -6],
           [ 9, -9],[-21, -9],[6,0], [9, 3],
           [ 9,9], [24, -6], [-3, 3], [-9, 3]);
  sub p {
    my ($k) = @_;
    return @{$p[$k % scalar(@p)]};
  }
}
sub bk {
  my ($k) = @_;
  my ($x,$y) = (2,0);
  foreach (1 .. $k) {
    ($x,$y) = mul($x,$y);
  }
  return ($x,$y);
}
sub P1 {
  my ($k) = @_;
  my ($x,$y) = bk($k);
  my ($px,$py) = p($k);
  return (($x + $px)/-24, ($y + $py)/-24);
}

sub P2 {
  my ($k) = @_;
  my ($x,$y) = bk($k+1);
  my ($px,$py) = p($k+1);
  return (($x + $px)/-24, ($y + $py)/-24);
}
sub P3 {
  my ($k) = @_;
  my ($x,$y) = bk($k+2);
  my ($px,$py) = p($k+2);
  return (($x + $px)/-24, ($y + $py)/-24);
}
sub P4 {
  my ($k) = @_;
  my ($x,$y) = bk($k+3);
  my ($px,$py) = p($k+3);
  return (($x + $px)/-24, ($y + $py)/-24);
}
sub P5 {
  my ($k) = @_;
  my ($x,$y) = bk($k+4);
  my ($px,$py) = p($k+4);
  return (($x + $px)/-24, ($y + $py)/-24);
}
sub P6 {
  my ($k) = @_;
  my ($x,$y) = bk($k+5);
  my ($px,$py) = p($k+5);
  return (($x + $px)/-24, ($y + $py)/-24);
}

sub P7 {
  my ($k) = @_;
  my ($x,$y) = bk($k+5);
  my ($rx,$ry) = xy_rotate_plus60($x,$y);
  my ($px,$py) = p($k+6);
  return ((9*$x + $rx + 9*$px)/-216, (9*$y + $ry + 9*$py)/-216);
}


sub hull_expand_RV {
  my ($points_R, $points_V, $ex,$ey, %options) = @_;
  ### hull_expand(): "end=$ex,$ey  R " . points_str($points_R) ."  V " . points_str($points_V)

  #                *---*
  #  *---*  =>      \   V+90deg
  #              *---*
  #                R
  #
  my $new_points_R = [ @$points_R,

                       (map{ my ($x,$y) = @$_;
                             ($x,$y) = xy_rotate_plus60($x,$y);
                             $x += $ex;
                             $y += $ey;
                             [ $x,$y ]}
                        @$points_V),
                     ];

  #                 *
  #                / \
  #    *          *   * V
  #   / \   =>   /
  #  *   *      * R+60deg
  #
  my ($ex60,$ey60) = xy_rotate_plus60($ex,$ey);
  my $new_points_V = [ (map{ my ($x,$y) = @$_;
                             [ xy_rotate_plus60($x,$y) ]}
                        @$points_R),

                       (map{ my ($x,$y) = @$_;
                             $x += $ex60;
                             $y += $ey60;
                             [ $x,$y ]}
                        @$points_V),
                     ];

  $points_R = $new_points_R;
  $points_V = $new_points_V;

  ($ex,$ey) = mul($ex,$ey);

  # print "% at exy=$ex,$ey\n";
  # print "%   R  ",points_str($points_R),"\n";
  # print "%   V  ",points_str($points_V),"\n";

  require Math::Geometry::Planar;
  my $planar = Math::Geometry::Planar->new;
  foreach my $points ($points_R, $points_V) {  # mutate
    $planar->points($points);
    if (@$points >= 5) {
      ### convexhull2 of: points_str($points)
      $planar = $planar->convexhull2;
      $points = $planar->points;
    }
    unless ($options{'unsorted'}) {
      $points = hull_points_sort_by_angle($points,
                                          xy_rotate_180($ex,$ey));
    }
  }
  return ($points_R, $points_V, $ex,$ey);
}

#------------------------------------------------------------------------------
# Alternate Terdragon

# return $x,$y unfolded around pivot $ex,$ey
sub unfold_middle_down {
  my ($x,$y, $ex,$ey) = @_;
  ($x,$y) = xy_rotate_minus120($x,$y);
  return ($ex+$x, $ey+$y);
}
sub unfold_end_down {
  my ($x,$y, $ex,$ey) = @_;
  my ($mx,$my) = xy_rotate_minus60($ex,$ey);
  return ($mx+$x, $my+$y);
}

sub AltEnd {
  my ($k) = @_;
  my $x = 2*3**int($k/2);  # inherit bignum
  my $y = $x * 0;
  if ($k%2) { ($x,$y) = mul($x,$y); }
  return ($x,$y);
}

sub alternate_hull_expand {
  my ($points, $ex,$ey, %options) = @_;
  ### hull_expand(): "end=$ex,$ey  " . points_str($points)

  my $down = ($ey != 0);
  my $middle_func = ($down ? \&unfold_middle_down : \&unfold_middle);
  my $end_func    = ($down ? \&unfold_end_down    : \&unfold_end);
  # print " % at pxy=$ex,$ey  ",points_str($points),"\n";
  $points = [ @$points,
              (map{[ $middle_func->(@$_, $ex,$ey) ]} @$points),
              (map{[ $end_func   ->(@$_, $ex,$ey) ]} @$points),
            ];

  require Math::Geometry::Planar;
  my $planar = Math::Geometry::Planar->new;
  $planar->points($points);
  if (@$points >= 5) {
    ### convexhull2 of: points_str($points)
    $planar = $planar->convexhull2;
    $points = $planar->points;
  }
  unless ($options{'unsorted'}) {
    $points = hull_points_sort_by_angle($points,
                                        xy_rotate_180($ex,$ey));
  }
  ($ex,$ey) = mul($ex,-$ey);
  return ($points, $ex,$ey);
}

sub alternate_hull_and_end {
  my ($k) = @_;
  my $ex = 2;
  my $ey = 0;
  my $points = [ [0,0], [$ex,$ey] ];
  foreach (1 .. $k) {
    ($points,$ex,$ey) = alternate_hull_expand($points,$ex,$ey);
  }
  return ($points,$ex,$ey);
}


#------------------------------------------------------------------------------

sub path_tree_depth_to_row_hull_area {
  my ($path, $depth) = @_;
  my $points = [ ];
  my ($n_row, $n_end) = $path->tree_depth_to_n_range($depth);
  foreach my $n ($n_row .. $n_end) {
    my ($x,$y) = $path->n_to_xy($n);
    push @$points, [$x,$y];
  }
  if (@$points < 3) {
    return 0;
  }
  return points_to_hull_area($points);
}

sub path_tree_depth_to_hull_area {
  my ($path, $depth) = @_;
  my $points = [ [1,1] ];
  foreach my $n ($path->n_start .. $path->tree_depth_to_n_end($depth)) {
    my ($x,$y) = $path->n_to_xy($n);
    push @$points, [$x,$y];
  }
  return points_to_hull_area($points);
}

sub points_to_hull_area {
  my ($points) = @_;
  if (@$points < 3) {
    return 0;
  }
  require Math::Geometry::Planar;
  my $planar = Math::Geometry::Planar->new;
  $planar->points($points);
  if (@$points >= 5) { $planar = $planar->convexhull2; }
  return $planar->area;
}

#------------------------------------------------------------------------------

sub dxdy_to_dir3_maybe {
  my ($dx,$dy) = @_;
  if ($dx == 2 && $dy == 0) {
    return 0;
  }
  if ($dx == -1) {
    if ($dy == 1) {
      return 1;
    }
    if ($dy == -1) {
      return 2;
    }
  }
  return undef;
}

sub dxdy_to_dir3 {
  my ($dx,$dy) = @_;
  return dxdy_to_dir3_maybe($dx,$dy)
    // croak "dxdy_to_dir3() bad $dx, $dy";
}

sub dxdy_to_dir6 {
  my ($dx,$dy) = @_;
  if ($dy == 0) {
    if ($dx == 2) { return 0; }
    if ($dx == -2) { return 3; }
  }
  if ($dy == 1) {
    if ($dx == 1) { return 1; }
    if ($dx == -1) { return 2; }
  }
  if ($dy == -1) {
    if ($dx == 1) { return 5; }
    if ($dx == -1) { return 4; }
  }
  die "unrecognised $dx,$dy";
}

#------------------------------------------------------------------------------
# p   =    X      + sqrt(3)*Y*i
# w12 = sqrt(3)/2 + 1/2*i
# p*w12 = (X*sqrt(3)/2 - sqrt(3)*Y*1/2)
#       + (X/2 + Y*sqrt(3)*sqrt(3)/2)*i
#       = (X*sqrt(3)/2 - Y*sqrt(3)/2)
#       + (X/sqrt(3)/2 + Y*sqrt(3)/2)*sqrt(3)*i

sub xy_rotate_plus30 {
  my ($x, $y) = @_;
  return ($x*(sqrt(3)/2) - $y*(sqrt(3)/2),
          $x/(2*sqrt(3)) + $y*(sqrt(3)/2));
}
sub xy_rotate_minus30 {
  my ($x, $y) = @_;
  return ($x*(sqrt(3)/2) + $y*(sqrt(3)/2),
          $x/(-2*sqrt(3)) + $y*(sqrt(3)/2));
}

sub xy_rotate_plus60 {
  my ($x, $y) = @_;
  return (($x-3*$y)/2,  # rotate +60
          ($x+$y)/2);
}
sub xy_rotate_minus60 {
  my ($x, $y) = @_;
  return (($x+3*$y)/2,  # rotate -60
          ($y-$x)/2);
}

sub xy_rotate_plus120 {
  my ($x,$y, $reps) = @_;
  if (defined $reps) {
    $reps %= 3;
  } else {
    $reps = 1;
  }
  foreach (1 .. $reps) {
    ($x,$y) = (($x+3*$y)/-2,  # rotate +120
               ($x-$y)/2);
  }
  return ($x,$y);
}
sub xy_rotate_minus120 {
  my ($x, $y) = @_;
  return ((3*$y-$x)/2,              # rotate -120
          ($x+$y)/-2);
}

sub xy_rotate_plus90 {
  my ($x,$y) = @_;
  return (-$y,$x);  # rotate +90
}
sub xy_rotate_minus90 {
  my ($x,$y) = @_;
  return ($y,-$x);  # rotate -90
}

sub xy_rotate_180 {
  my ($x, $y) = @_;
  return (-$x, -$y);
}

sub BitAboveLowestOne {
  my ($n) = @_;
  for (;;) {
    if (! $n || ($n % 2) != 0) {
      last;
    }
    $n = int($n/2);
  }
  $n = int($n/2);
  return ($n % 2);
}

#------------------------------------------------------------------------------
# dir(), turn()

sub dir {
  my ($n) = @_;
  $n = cnv($n,10,3);
  return scalar($n =~ tr/1/1/);
}
sub turn {
  my ($n) = @_;
  until ($n % 3) {
    $n /= 3;
    $n >= 1 || croak "turn() is for n>=1";
  }
  return ($n % 3 == 1 ? 1 : -1);
}

sub TurnLpred {
  my ($n) = @_;
  $n >= 0 || croak "TurnLpred() is for n>=0";
  return $n>=1 && turn($n) == 1 ? 1 : 0;
}
sub TurnRpred {
  my ($n) = @_;
  $n >= 0 || croak "TurnLpred() is for n>=0";
  return $n>=1 && turn($n) == -1 ? 1 : 0;
}
use constant::defer TurnLpred_flat => sub {
  require FLAT;
  require MyFLAT;
  return FLAT::Regex->new('(0|1|2)* 1 0*')->as_dfa  # lowest non-0 is 1
    ->MyFLAT::set_name('TurnLpred');
};
use constant::defer TurnRpred_flat => sub {
  require MyFLAT;
  return FLAT::Regex->new('(0|1|2)* 2 0*')->as_dfa  # lowest non-0 is 2
    ->MyFLAT::set_name('TurnRpred');
};


#------------------------------------------------------------------------------
# dNextL, dNextR

sub dNextL {
  my ($n) = @_;
  my $inc = ($n%3 == 1 ? 2 : 1);
  return $inc + TurnRpred($n+$inc);
}
sub dNextR {
  my ($n) = @_;
  return ($n%3 ? 2+turn($n+1) : 2);
}

our @dNextL_state_table = ([0,1,0],   # 0
                           [0,2,1],   # 1
                           [0,2,1],   # 2
                          );
our @dNextL_state_to_d = (1,2,3);
sub dNextL_flat {
  my ($d) = @_;
  require MyFLAT;
  return MyFLAT::aref_to_FLAT_DFA(\@dNextL_state_table,
                                  name => "dNextL=$d",
                                  accepting => $d-1);
}
memoize('dNextL_flat', LIST_CACHE=>'MERGE');

our @dNextR_state_table = ([1,0,0],   # 0
                           [1,0,2],   # 1
                           [1,0,2],   # 2
                          );
our @dNextR_state_to_d = (1,2,3);
sub dNextR_flat {
  my ($d) = @_;
  require MyFLAT;
  return MyFLAT::aref_to_FLAT_DFA(\@dNextR_state_table,
                                  name => "dNextR=$d",
                                  starting => 1,  # state d2
                                  accepting => $d-1);
}
memoize('dNextR_flat', LIST_CACHE=>'MERGE');

our @dNextLsub1_state_table = ([2,1,0],   # 0 d1
                               [4,3,1],   # 1 d1b
                               [2,1,0],   # 2 d2  (start)
                               [4,3,1],   # 3 d2b
                               [2,1,0]);  # 4 d3
our $dNextLsub1_start_state = 2; # d2
our @dNextLsub1_state_to_name = ('d1','d1b','d2','d2b','d3');
our @dNextLsub1_state_to_d = (1,1,2,2,3);
sub dNextLsub1_flat {
  my ($d) = @_;
  require MyFLAT;
  return MyFLAT::aref_to_FLAT_DFA
    (\@dNextLsub1_state_table,
     name => "dNextLsub1=$d",
     starting => $dNextLsub1_start_state,
     accepting_list => [grep {$dNextLsub1_state_to_d[$_]==$d}
                        0 .. $#dNextLsub1_state_table]);
}
memoize('dNextLsub1_flat', LIST_CACHE=>'MERGE');


#------------------------------------------------------------------------------
# Rpred

sub Rpred_of_ternary_str {
  my ($str) = @_;
  return ! ($str =~ /11|12|20/);
}
sub Rpred {
  my ($n) = @_;
  return Rpred_of_ternary_str(cnv($n,10,3));
}

our @Rpred_state_table = ([0,1,2],   # R
                          [0,3,3],   # M
                          [3,1,2],   # E
                          [3,3,3],   # non
                         );
use constant::defer NonRpred_flat => sub {
  require MyFLAT;
  MyFLAT::aref_to_FLAT_DFA(\@Rpred_state_table, name => 'NonRpred');
};
use constant::defer Rpred_flat => sub {
  return NonRpred_flat()->complement
    ->MyFLAT::set_name('Rpred');
};

our $Rpred_regexp = '((2*1)?0)*2*1?';
our $Rpred_LtoH_regexp = '1?2*(0(12*)?)*';
our $NonRpred_regexp = '11|12|20';
# our $NonRpred_regexp = '[012]*(11|12|20)[012]*';

#------------------------------------------------------------------------------
# RpointPred

sub RpointPred {
  my ($n) = @_;
  foreach my $n (TerdragonCurve_path()->n_to_n_list($n)) {
    if (Rpred($n)) { return 1; }
  }
  return 0;
}
our $RpointPred_regexp
  = qr/(0|10|(2+10))*       # back to R
       (                    # stop R
        |2+(022+)?          # stop E or on to E4
        |2*1                # stop ME or M
        |12                 # stop T
        |(11|2+11)0*        # stop M2
         (1?2)?             # to T
        )
      /x;

our $RpointPred_enclosures = MyTerdragonConf->new
  (name => 'RpointPred enclosures',
   segs => [
            '-1,-1,1',
            '-1,1,0',
            '1,1,0',
            # '4,0,2',
           ]);

our $RpointPred_around = MyTerdragonConf->new
  (name => 'RpointPred around',
   segs => [
            '0,0,0',      # around start
            '0,0,1',
            '0,0,2',
            '1,1,2',
            '1,-1,1',
            '-2,0,0',

            # '2,0,0',      # around end
            '2,0,1',
            '2,0,2',
            '3,1,2',
            # '3,-1,1',
           ]);

our $RpointPred_segs
  = $RpointPred_around->union($RpointPred_enclosures)
  ->set_name('RpointPred segs'),
# $RpointPred_segs->print;

our @RpointPred_state_to_conf;
our @RpointPred_state_table;
our ($NonRpointPred_flat, $RpointPred_flat);

{
  my %RpointPred_conf_to_state;
  sub RpointPred_conf_to_state {
    my ($conf) = @_;
    my $key = $conf->key;
    my $state = $RpointPred_conf_to_state{$key};
    if (! defined $state) {
      push @RpointPred_state_to_conf, $conf;
      $state = $#RpointPred_state_to_conf;
      $RpointPred_conf_to_state{$key} = $state;
      # print "new state $state\n";
    }
    return $state;
  }
}

our @RpointPred_R_triangles
  = (MyTerdragonConf->new(name => 'RpointPred R top left',
                       segs => [ '-1,1,0', '0,0,1', '1,1,2' ]),
     MyTerdragonConf->new(name => 'RpointPred R bottom left',
                       segs => [ '-2,0,0', '-1,-1,1', '0,0,2' ]),
     MyTerdragonConf->new(name => 'RpointPred R bottom middle',
                       segs => [ '0,0,0', '1,-1,1', '2,0,2' ]),
    );
# foreach my $conf (@RpointPred_R_triangles){$conf->print;}

# A nearby triangle is boundary of respective $side when it has at least one
# segment $conf but not all of its segments in conf.
sub RpointPred_is_boundary {
  my ($conf) = @_;
  foreach my $triangle (@RpointPred_R_triangles) {
    my $got = $triangle->intersection ($conf);
    if ($got->num_segs == 0) {
      # no segments at all around this square, not boundary
      next;
    }
    my $all = $triangle->intersection ($conf->{'within'});
    if ($got->equals($all)) {
      # all segments (of "within") around this square, not boundary
      next;
    }
    # at least 1 but fewer than all segments, so boundary
    return 1;
  }
  return 0;
}

use constant::defer RpointPred_init => sub {
  my %options = @_;
  my $verbose = $options{'verbose'} || 0;
  my $verbose_print = $options{'verbose_print'} || sub { print @_; };

  require MyFLAT;
  my @state_enclosed_list;
  my @RpointPred_state_is_R_enclosed;

  RpointPred_conf_to_state(MyTerdragonConf->initial
                           (within => $RpointPred_segs));
  {
    my $state = 0;
    my $depth = 0;
    for (;;) {
      $depth++;
      my $last_state = $#RpointPred_state_to_conf;
      for ( ; $state <= $last_state; $state++) {
        my $from_conf = $RpointPred_state_to_conf[$state];
        $from_conf->set_name("RpointPred state $state");
        if ($verbose >= 3) {
          $verbose_print->("########################################\n");
          $verbose_print->("from $state  ",$from_conf->key,"\n");
          $verbose_print->($from_conf->print_to_str);
        }
        foreach my $digit (0,1,2) {
          my $to_conf = $from_conf->expand($digit);
          my $to_state = RpointPred_conf_to_state($to_conf);
          $RpointPred_state_table[$state]->[$digit] = $to_state;

          if ($verbose >= 3) {
            $verbose_print->("  digit=$digit to $RpointPred_state_table[$state]->[$digit]  ",$to_conf->key,"\n");
            $verbose_print->($to_conf->print_to_str);
          }
        }
      }
      my $new = $#RpointPred_state_to_conf - $last_state;
      if ($verbose) {
        $verbose_print->("depth $depth new $new\n");
      }
      last if $new == 0;
    }
  }

  foreach my $i (0 .. $#RpointPred_state_to_conf) {
    my $conf = $RpointPred_state_to_conf[$i];
    $RpointPred_state_is_R_enclosed[$i] = ! RpointPred_is_boundary($conf);
  }
  @state_enclosed_list
    = grep {$RpointPred_state_is_R_enclosed[$_]} 0..$#RpointPred_state_table;
  if ($verbose) {
    $verbose_print->("RpointPred enclosed counts ",
                     scalar @state_enclosed_list);
  }

  if ($verbose) {
    $verbose_print->("  enclosed R = ",join(',',@state_enclosed_list),
                     " count ",scalar(@state_enclosed_list),"\n");
  }

  foreach my $i (@state_enclosed_list) {
    foreach my $digit (0,1,2) {
      my $to = $RpointPred_state_table[$i]->[$digit];
      unless ($RpointPred_state_is_R_enclosed[$to]) {
        # if ($verbose) {
        #   $verbose_print->("enclosed $i digit=$digit didn't remain so\n");
        # }
        # $RpointPred_state_to_conf[$i]->print;
        # $RpointPred_state_to_conf[$i]->expand($digit)->print;
        # die;
      }
    }
  }

  {
    # R
    my $f = MyFLAT::aref_to_FLAT_DFA(\@RpointPred_state_table,
                                     name => 'NonRpointPred',
                                     accepting_list => \@state_enclosed_list,
                                    );
    $NonRpointPred_flat = $f;
    if ($verbose >= 2) {
      $verbose_print->("$f->{name} accepting ",join(' ',$f->get_accepting),"\n");
    }
    $f = $f->complement->MyFLAT::set_name('RpointPred');
    $RpointPred_flat = $f;
  }
};



#------------------------------------------------------------------------------
# Lpred

sub Lpred_k_of_ternary_str {
  my ($str) = @_;
  return scalar($str !~ /02|10|11/);
}
sub cnv_padded {
  my ($n, $from, $to, $width) = @_;
  $n = cnv($n,$from,$to);
  if ((my $zeros = $width - length($n)) > 0) {
    $n = ('0'x $zeros) . $n;
  }
  return $n;
}
sub Lpred_k {
  my ($n,$k) = @_;
  $n = cnv_padded($n,10,3,$k);
  return (length($n) == $k && Lpred_k_of_ternary_str($n));
}

# digit reversal 0<->2
our @Lpred_state_table = map {[reverse @$_]} @Rpred_state_table;
use constant::defer NonLpred_k_flat => sub {
  require MyFLAT;
  MyFLAT::aref_to_FLAT_DFA(\@Lpred_state_table, name => 'NonLpred_k');
};
use constant::defer Lpred_k_flat => sub {
  return NonLpred_k_flat()->complement
    ->MyFLAT::set_name('Lpred_k');
};

#------------------------------------------------------------------------------
# Bpred

# Return a FLAT::DFA which matches boundary segment numbers.
# A string of $k many bits (high to low) is accepted if that N is a boundary
# segment (left or right or both).
use constant::defer Bpred_k_flat => sub {
  # union() gives an NFA, want to go to DFA
  return Rpred_flat()->union(Lpred_k_flat())->as_dfa
    ->MyFLAT::set_name('Bpred');
};

# Return a FLAT::DFA which matches non-boundary segment numbers.
# A string of $k many bits (high to low) is accepted if that N is a
# non-boundary segment (neither left nor right boundary).
use constant::defer NonBpred_k_flat => sub {
  # both NonR and NonL, so the intersection of those sets
  return NonRpred_flat()->intersect(NonLpred_k_flat())
    ->MyFLAT::set_name('NonBpred');
};

#------------------------------------------------------------------------------
# Dpred

sub Tperm {
  my ($n) = @_;
  my $n3 = cnv($n,10,3);
  $n3 =~ s{([12])0}{($1==1?2:1).'0'}eg;
  return cnv($n3,3,10);
}

# ternary digits high to low which are a segment with a 1-sided boundary
# triangle on its right
use constant::defer R1sides_flat => sub {
  require MyFLAT;
  return MyFLAT::aref_to_FLAT_DFA(\@Rpred_state_table,
                                  accepting_list => [0]) # R
    ->MyFLAT::set_name('R1sides');
};

# ternary digits high to low which, within the given k many digits, are a
# segment with a 1-sided boundary triangle on its left
use constant::defer L1sides_k_flat => sub {
  require MyFLAT;
  return MyFLAT::aref_to_FLAT_DFA(\@Lpred_state_table,
                                  accepting_list => [0]) # L
    ->MyFLAT::set_name('L1sides k');
};

use constant::defer R2sides_flat => sub {
  require MyFLAT;
  return MyFLAT::aref_to_FLAT_DFA(\@Rpred_state_table,
                                  accepting_list => [1,2]) # M,E
    ->MyFLAT::set_name('R2sides');
};
use constant::defer L2sides_k_flat => sub {
  require MyFLAT;
  return MyFLAT::aref_to_FLAT_DFA(\@Lpred_state_table,
                                  accepting_list => [1,2]) # M,E
    ->MyFLAT::set_name('L2sides k');
};

# ternary digits high to low which are a single-visited point, with left
# turn or right turn or either
use constant::defer SpredLeft_flat => sub {
  require MyFLAT;
  return R1sides_flat()->MyFLAT::concat(FLAT::Regex->new('1 0*'))
    ->MyFLAT::set_name('SpredLeft k');
};
use constant::defer SpredRight_k_flat => sub {
  require MyFLAT;
  return L1sides_k_flat()->MyFLAT::concat(FLAT::Regex->new('2 0*'))
    ->MyFLAT::set_name('SpredRight k');
};
use constant::defer Spred_k_flat => sub {
  require MyFLAT;
  return SpredLeft_flat()->union(SpredRight_k_flat(),
                                 FLAT::Regex->new('0*')->as_dfa)
    ->as_dfa
    ->MyFLAT::set_name('Spred k');
};

# ternary digits high to low which are a double-visited point, with left
# turn or right turn or either
use constant::defer DpredLeft_flat => sub {
  require MyFLAT;
  return R2sides_flat()->MyFLAT::concat(FLAT::Regex->new('1 0*'))
    ->MyFLAT::set_name('DpredLeft k');
};
use constant::defer DpredRight_k_flat => sub {
  require MyFLAT;
  return L2sides_k_flat()->MyFLAT::concat(FLAT::Regex->new('2 0*'))
    ->MyFLAT::set_name('DpredRight k');
};
use constant::defer Dpred_k_flat => sub {
  require MyFLAT;
  return DpredLeft_flat()->union(DpredRight_k_flat())->as_dfa
    ->MyFLAT::set_name('Dpred k');
};

# ternary digits high to low which are a triple-visited point, with left
# turn or right turn or either
use constant::defer TpredLeft_flat => sub {
  require MyFLAT;
  return NonRpred_flat()->MyFLAT::concat(FLAT::Regex->new('1 0*'))
    ->MyFLAT::set_name('TpredLeft k');
};
use constant::defer TpredRight_k_flat => sub {
  require MyFLAT;
  return NonLpred_k_flat()->MyFLAT::concat(FLAT::Regex->new('2 0*'))
    ->MyFLAT::set_name('TpredRight k');
};
use constant::defer Tpred_k_flat => sub {
  require MyFLAT;
  return TpredLeft_flat()->union(TpredRight_k_flat())->as_dfa
    ->MyFLAT::set_name('Tpred k');
};


#------------------------------------------------------------------------------
# EpredR, EpredL

our @EpredR_state_table = ([0,0,1],   # rs0
                           [2,0,1],   # rs1
                           [2,1,2],   # rs2, enclosed
                          );
use constant::defer RsideNum1_flat => sub {
  require MyFLAT;
  MyFLAT::aref_to_FLAT_DFA(\@EpredR_state_table, name => 'ER1sides',
                           accepting_list=>[0]);
};
use constant::defer RsideNum2_flat => sub {
  require MyFLAT;
  MyFLAT::aref_to_FLAT_DFA(\@EpredR_state_table, name => 'ER2sides',
                           accepting_list=>[1]);
};
use constant::defer EpredR_flat => sub {
  require MyFLAT;
  MyFLAT::aref_to_FLAT_DFA(\@EpredR_state_table, name => 'EpredR');
};

sub EpredL {
  my ($n) = @_;
  return EpredL_of_ternary_str(cnv($n,10,3));
}
sub EpredL_of_ternary_str {
  my ($n3) = @_;
  if ($n3 =~ /0$/) { return 0; }
  if ($n3 =~ /.*(10|11)/) {
    $n3 = substr($n3,$+[0]);
    if ($n3 =~ /02/) { return 0; }
    return 1;
  }
  return 0;
}

our @EpredL_state_table = ([0,1,0],   # ls0
                           [1,2,0],   # ls1
                           [1,2,2],   # ls2
                          );
use constant::defer LsideNum1_flat => sub {
  require MyFLAT;
  MyFLAT::aref_to_FLAT_DFA(\@EpredL_state_table, name => 'EL1sides',
                           accepting_list=>[0]);
};
use constant::defer LsideNum2_flat => sub {
  require MyFLAT;
  MyFLAT::aref_to_FLAT_DFA(\@EpredL_state_table, name => 'EL2sides',
                           accepting_list=>[1]);
};
use constant::defer EpredL_flat => sub {
  require MyFLAT;
  MyFLAT::aref_to_FLAT_DFA(\@EpredL_state_table, name => 'EpredL');
};

use constant::defer Epred_flat => sub {
  require MyFLAT;
  return EpredL_flat()->union(EpredR_flat())
    ->MyFLAT::set_name('Epred');
};

use constant::defer EpredRpair_flat => sub {
  require MyFLAT;
  return EpredR_flat()
    ->intersect(EpredR_flat()->MyFLAT::digits_increment(add => -1))
    ->MyFLAT::minimize
    ->MyFLAT::set_name('EpredRpair');
};
use constant::defer EpredLpair_flat => sub {
  require MyFLAT;
  return EpredL_flat()
    ->intersect(EpredL_flat()->MyFLAT::digits_increment(add => -1))
    ->MyFLAT::minimize
    ->MyFLAT::set_name('EpredLpair');
};

our @EpredTwelve_state_table = ([ 7,1,7 ],   # 0 = t1
                                [ 7,2,7 ],   #     t2
                                [ 7,7,3 ],   #     t3
                                [ 7,4,3 ],   #     t4
                                [ 7,5,4 ],   #     t5
                                [ 5,6,4 ],   #     t6
                                [ 6,6,6 ],   # 6 = enc12
                                [ 7,7,7 ]);  # 7 = non

use constant::defer EpredTwelve_LtoH_flat => sub {
  require MyFLAT;
  MyFLAT::aref_to_FLAT_DFA(\@EpredTwelve_state_table,
                           accepting => 6,
                           name => 'EpredTwelve LtoH');
};

#------------------------------------------------------------------------------

sub Visits_k_flat {
  my ($num) = @_;
  return ($num == 1 ? Spred_k_flat()
          : $num == 2 ? Dpred_k_flat()
          : $num == 3 ? Tpred_k_flat()
          : die);
}

use constant::defer fExact3k_flat => sub {
  require MyFLAT;
  return FLAT::Regex->new('(0|1|2)* (0* | 2*)')
    ->MyFLAT::minimize
    ->MyFLAT::set_name('fExact3k');
};

sub fVisits_nonexact_2_flat {
  require MyFLAT;
  my $f = FLAT::Regex->new('(0|1|2)*')->as_dfa
    ->MyFLAT::concat(Bpred_k_flat());             # fBpred in sub-curve
  return $f->MyFLAT::minimize
    ->MyFLAT::set_name('fVisits_nonexact_2');

  # FIXME: Was this right?
  #
  # return NonBpred_k_flat()->intersect($f)         # but fNonBpred whole
  #   ->MyFLAT::minimize
  #   ->MyFLAT::set_name('fVisits_nonexact_2');
}

# fVisits = 1 always possible by infinitely many 02,10,11 and infinitely
# many 11,12,20 so not on sub-curve boundary.  So at k many digits can never
# exclude fVisits=1.
#
sub fVisits_exact_flat {
  my ($num) = @_;
  require MyFLAT;
  my $f = Visits_k_flat($num)
    ->MyFLAT::fraction_also_nines
    ->MyFLAT::minimize
    ->MyFLAT::set_name("fVisits exact $num");

}
memoize('fVisits_flat', LIST_CACHE=>'MERGE');

sub fVisits_flat {
  my ($num) = @_;
  require MyFLAT;
  my $f = Visits_k_flat($num);
  my $exact_0 = $f->MyFLAT::concat(FLAT::Regex->new('0*'));
  my $exact_2 = $f->MyFLAT::digits_increment(add => -1)
    ->MyFLAT::concat(FLAT::Regex->new('2*'));
  $f = $exact_0->union($exact_2);
  $f = $f->MyFLAT::minimize;

  if ($num == 2) {
    $f = $f->union(fVisits_nonexact_2_flat());
  }
  if ($num == 1) {
    $f = $f->union(FLAT::Regex->new('(0|1|2)*')->as_dfa
                   ->MyFLAT::concat(fNonBpred_flat()));
    $f = $f->union(fBpred_flat());
  }
  $f = $f->MyFLAT::minimize;
  return $f->MyFLAT::set_name("fVisits $num");

}
memoize('fVisits_flat', LIST_CACHE=>'MERGE');

#------------------------------------------------------------------------------
# MyTerdragonConf is an object representing a given unit line segment and
# possible surrounding segments.
#
# "within" is another MyTerdragonConf which is segments the configuration is
# supposed to be a subset of.  The expand() method prunes its output to just
# those of "within".

{
  package MyTerdragonConf;
  use Carp 'croak';
  use List::Util 'min','max';
  use List::MoreUtils;

  sub new {
    my $class = shift;   # ($class, segs=>$s)
    if (ref $class) { $class = ref $class; }
    my $self = bless { @_ }, $class;
    $self->{'segshash'} = { map {$_=>1} $self->segs_list };
    foreach my $seg ($self->segs_list) {
      $self->seg_validate($seg);
    }
    return $self;
  }
  sub clone {
    my $self = shift;   # ($class, segs=>$s)
    return $self->new (segs => [$self->segs_list],
                       @_);
  }
  sub seg_validate {
    my ($self, $seg) = @_;
    unless ($seg =~ /^(-?[1-9]|[0-9])[0-9]*,(-?[1-9]|[0-9])[0-9]*,[012]$/) {
      croak "Bad seg numbers: ",$seg;
    }
    my ($x,$y,$dir) = split /,/, $seg;
    if (($x+$y)&1) {
      croak "Bad seg, X+Y is odd: $x, $y";
    }
  }
  sub key {
    my ($self) = @_;
    return ($self->{'key'} //= join(' ',sort $self->segs_list));
  }
  sub initial {
    my $class = shift;
    return $class->new (segs => ['0,0,0'], @_);
  }
  sub equals {
    my ($self, $other) = @_;
    return $self->key eq $other->key;
  }
  sub contains {
    my ($self, $seg) = @_;
    # $self->seg_validate($seg);
    return $self->{'segshash'}->{$seg};
  }
  sub is_subset {
    my ($self, $superset) = @_;
    foreach my $seg ($self->segs_list) {
      if (! $superset->contains($seg)) { return 0; }
    }
    return 1;
  }
  sub segs_list {
    my ($self) = @_;
    return @{$self->{'segs'}};
  }
  sub num_segs {
    my ($self) = @_;
    return scalar @{$self->{'segs'}};
  }

  # return a new MyTerdragonConf with each point $x,$y moved to $x+$dx, $y+$dy
  sub translate {
    my ($self, $dx,$dy) = @_;
    if ($dx==0 && $dy==0) { return $self; }
    my $name = $self->{'name'};
    if (defined $name) {
      $name = "$name moved $dx,$dy";
    }
    return $self = $self->new
      (name => $name,
       segs => [ map { my ($x,$y,$dir) = split /,/;
                       $x += $dx;
                       $y += $dy;
                       "$x,$y,$dir"
                     } $self->segs_list ]);
  }

  # return a new MyTerdragonConf with each segment rotated by $ddir, so each
  # $dir becomes $dir+$ddir, mod 3
  sub rotate {
    my ($self, $ddir) = @_;
    $ddir %= 4;
    if ($ddir == 0) { return $self; }

    my $name = $self->{'name'};
    if (defined $name) {
      $name = "$name rotate ".($ddir==1   ? "+120"
                               : $ddir==2 ? "-120"
                               : $ddir);
    }
    return $self = $self->new
      (name => $name,
       segs => [ map { my ($x,$y,$dir) = split /,/;
                       foreach (1 .. $ddir) {
                         ($x,$y) = MyTerdragon::xy_rotate_plus120($x,$y)
                       }
                       $dir = ($dir + $ddir) % 3;
                       "$x,$y,$dir"
                     } $self->segs_list ]);
  }

  # return a new MyTerdragonConf with the segment common to all the given confs
  sub intersection {
    my $self = shift;
    foreach my $other (@_) {
      $self = $self->new
        (within => $self->{'within'},
         segs => [ grep {$other->contains($_)} $self->segs_list ]);
    }
    return $self;
  }
  sub difference {
    my $self = shift;
    foreach my $other (@_) {
      $self =$self->new
        (within => $self->{'within'},
         segs => [ grep {! $other->contains($_)} $self->segs_list ]);
    }
    return $self;
  }
  sub union {
    my $self = shift;
    my $name = $self->{'name'};
    if (defined $name) { $name = "$name union"; }
    return $self->new (name   => $name,
                       within => $self->{'within'},
                       segs   => [ List::MoreUtils::uniq
                                   (map {$_->segs_list} $self,@_) ]);

  }

  sub set_name {
    my ($self, $name) = @_;
    $self->{'name'} = $name;
    return $self;
  }
  sub set_within {
    my ($self, $within) = @_;
    $self->{'within'} = $within;
    return $self;
  }

  sub expand {
    my ($self, $digit, $alt) = @_;
    ($digit==0 || $digit==1 || $digit==2)
      or die "oops, epxand digit=$digit";
    # print "expand of: (alt=",$alt||0,")\n"; $self->print;

    # state 0 right side enc, state 1 left side enc
    #
    #
    #   x2,y2  * ---2---> *
    #            ^
    #             1
    #              \
    #  x,y * --0---> *  x1,y1        x,y * --0---> *  x1,y1
    #                                            /
    #                                           1
    #                                          v
    #                                 x2,y2  * ---2---> *
    #
    my @segs = map { my ($x,$y,$dir) = split /,/;
                     if ($alt) {
                       ($x,$y) = MyTerdragon::mul_conj($x,$y);
                     } else {
                       ($x,$y) = MyTerdragon::mul($x,$y);
                     }
                     my $x1 = $x + $dir3_to_dx[$dir];
                     my $y1 = $y + $dir3_to_dy[$dir];
                     my $dir1 = ($dir + ($alt ? -1 : 1)) % 3;  # of seg 1
                     my $x2 = $x1 + $dir3_to_dx[$dir1];
                     my $y2 = $y1 + $dir3_to_dy[$dir1];
                     ("$x,$y,$dir",
                      "$x1,$y1,$dir1",
                      "$x2,$y2,$dir")
                   } $self->segs_list;
    if ($digit == 1) {
      @segs = map { my ($x,$y,$dir) = split /,/;
                    $x-=2;
                    if ($alt) {
                      ($x,$y) = MyTerdragon::xy_rotate_plus120($x,$y);
                    } else {
                      ($x,$y) = MyTerdragon::xy_rotate_minus120($x,$y);
                    }
                    $dir = ($alt ? $dir+1 : $dir-1) % 3;
                    "$x,$y,$dir"
                  } @segs;
    } elsif ($digit == 2) {
      @segs = map { my ($x,$y,$dir) = split /,/;
                    $x-=1; $y += ($alt ? 1 : -1);
                    "$x,$y,$dir"
                  } @segs;
    }
    my $expand = $self->new (segs => \@segs);
    if (defined (my $within = $self->{'within'})) {
      $expand = $expand->intersection($within);
      $expand->{'within'} = $within;
    }
    if (defined (my $name = $self->{'name'})) {
      unless ($name =~ /[0-9]$/) { $name .= ' '; }
      $expand->{'name'} = $name . '.' . ($alt ? 'a' : '') . $digit;
    }
    return $expand;
  }
  sub alt_expand {
    my ($self, $digit) = @_;
    $self->{'alt'} // die;
    my $alt = 1-$self->{'alt'};
    $self = $self->expand($digit,$alt);
    $self->{'alt'} = $alt;
    return $self;
  }
  sub alt_expand9 {
    my ($self, $digit) = @_;
    (0 <= $digit && $digit <= 8) or die;
    my $within = $self->{'within'};
    { local $self->{'within'} = undef;
      $self = $self->expand(int($digit/3),1);
      $self->{'within'} = $within;
    }
    return $self->expand($digit%3);
  }
  sub mirror_vertical {
    my ($self) = @_;
    my $name = $self->{'name'};
    if (defined $name) {
      $name = "$name mirror vertical";
    }
    return $self = $self->new
      (name => $name,
       segs => [ map { my ($x,$y,$dir) = split /,/;
                       $y = -$y;
                       if ($dir) { $dir = 3-$dir; }
                       "$x,$y,$dir"
                     } $self->segs_list ]);
  }

  # return true if the configuration has a boundary triangle side 0=right 1=left
  sub is_boundary {
    my ($self, $side) = @_;
    # $self->print;
    my $within = $self->{'within'} // croak 'No within';
    my $dd = ($side
              ?  1    # side=1 to left, dir+1
              : -1);  # side=0 to right, dir-1
    ### $dd

  SEG: foreach my $seg ($self->segs_list) {
      ### $seg
      my ($x,$y,$dir) = split /,/, $seg;

      #               dir
      #         x,y  -----> x1,y1
      #           ^       /
      #   dir+2dd  \     /  dir+dd
      #             \   v
      #              x2,y2
      #
      my $dir1 = ($dir + $dd) % 3;
      my $dir2 = ($dir + 2*$dd) % 3;
      my $x1 = $x + $dir3_to_dx[$dir];
      my $y1 = $y + $dir3_to_dy[$dir];
      my $x2 = $x1 + $dir3_to_dx[$dir1];
      my $y2 = $y1 + $dir3_to_dy[$dir1];

      foreach my $other_seg ("$x1,$y1,$dir1",
                             "$x2,$y2,$dir2") {
        ### $other_seg
        if (! $within->contains($other_seg)) {
          # presence of $other_seg is unspecified, this square cannot be
          # used to test for boundary
          ### not within ...
          next SEG;
        }
        if (! $self->contains($other_seg)) {
          ### absent, so boundary ...
          return 1;
        }
      }
    }
    return 0;
  }

  # bounding box around the segments
  sub bbox {
    my ($self) = @_;
    my $x_min=0;
    my $x_max=1;
    my $y_min=0;
    my $y_max=0;
    foreach my $seg ($self->segs_list) {
      my ($x,$y,$dir) = split /,/, $seg;
      foreach (0,1) {
        $x_min = min($x_min, $x); $x_max = max($x_max, $x);
        $y_min = min($y_min, $y); $y_max = max($y_max, $y);
        $x += $dir3_to_dx[$dir];
        $y += $dir3_to_dy[$dir];
      }
    }
    return ($x_min,$y_min, $x_max,$y_max);
  }

  #
  # *      *       *
  #  ^    / ^     / ^
  #  \   /   \   /   \
  #   \ v     \ v     \
  #    *------>*------>*
  #           / ^
  #          /   \
  #         v     \
  #        *       *
  sub print {
    my ($self) = @_;
    print $self->print_to_str, "\n";
  }
  sub print_to_str {
    my ($self) = @_;
    ### print() ...
    ### within: $self->{'within'}
    my $segshash = $self->{'segshash'};
    my $R = $self->{'within'} && $self->is_boundary(0);
    my $L = $self->{'within'} && $self->is_boundary(1);

    my $ret = "conf";
    if (defined $self->{'name'}) {
      $ret .= ' ' . $self->{'name'};
    }
    if ($self->{'within'}) {
      $ret .= ' boundary';
      $ret .= ($R ? " R" : " nonR");
      $ret .= ($L ? " L" : " nonL");
    }
    if (defined $self->{'alt'}) {
      $ret .= " alt $self->{'alt'}";
    }
    $ret .= "\n";

    # sort for consistency of output
    $ret .= "   " . join(' ',sort $self->segs_list) . "\n";

    my ($x_min,$y_min, $x_max,$y_max) = $self->bbox;
    ### $x_min
    ### $x_max
    ### $y_min
    ### $y_max
    #$x_min -= ($x_min&1);

    foreach my $y (reverse $y_min .. $y_max) {
      my $yp1 = $y+1;
      if ($y<$y_max) {
        # above verticals
        $ret .= "      ";
        foreach my $x ($x_min .. $x_max-1) {
          my $xp1 = $x+1;
          if (($x+$y)&1) {
            if ($self->contains("$xp1,$y,1"))    { $ret .= " ^  "; }
            else                                 { $ret .= "    "; }
          } else {
            if ($self->contains("$xp1,$yp1,2"))  { $ret .= "   /"; }
            else                                 { $ret .= "    "; }
          }
        }
        $ret .= "\n";

        $ret .= "      ";
        foreach my $x ($x_min .. $x_max-1) {
          my $xp1 = $x+1;
          if (($x+$y)&1) {
            if ($self->contains("$xp1,$y,1"))    { $ret .= "  \\ "; }
            else                                 { $ret .= "    "; }
          } else {
            if ($self->contains("$xp1,$yp1,2"))  { $ret .= "  / "; }
            else                                 { $ret .= "    "; }
          }
        }
        $ret .= "\n";

        $ret .= "      ";
        foreach my $x ($x_min .. $x_max-1) {
          my $xp1 = $x+1;
          if (($x+$y)&1) {
            if ($self->contains("$xp1,$y,1"))    { $ret .= "   \\"; }
            else                                 { $ret .= "    "; }
          } else {
            if ($self->contains("$xp1,$yp1,2"))  { $ret .= " v  "; }
            else                                 { $ret .= "    "; }
          }
        }
        $ret .= "\n";
      }

      # row
      $ret .= sprintf " %2d:  ", $y;
      if (($x_min+$y) & 1) { $ret .= "    "; }
      for (my $x = $x_min + (($x_min+$y)&1); $x <= $x_max; $x+=2) {
        my $xp1 = $x+1;
        # $ret .= "\nx=$x y=$y\n";
        $ret .= ($x==0 && $y==0 ? "O" : "*");
        if ($self->contains("$x,$y,0")) { $ret .= ($x==0&&$y==0
                                                   ? "======>"
                                                   : "------>"); }
        else                          { $ret .= "       "; }
      }
      $ret .= "\n";
    }
    $ret .= "\n";
    $ret =~ s/ +\n/\n/g; # delete trailing spaces
    return $ret;
  }

  # my $path = MyTerdragon::TerdragonCurve_path()
}

#------------------------------------------------------------------------------

#    0,0 ====> 2,0
#      ^       /
#       \     /
#        \   v
#         1,-1
#
our $R_triangle_segs = MyTerdragonConf->new
  (name => 'R triangle',
   segs => [ "0,0,0",       # below m
             "2,0,2",
             "1,-1,1" ]);

#         1,1
#        /   ^
#       /     \
#      v       \
#    0,0 ====> 2,0
#
our $L_triangle_segs = MyTerdragonConf->new
  (name => 'L triangle',
   segs => [ "0,0,0",      # above m
             "2,0,1",
             "1,1,2" ]);

our $B_triangles_segs = $R_triangle_segs->union($L_triangle_segs)
  ->set_name('B two triangles');  # Bpred enclosing


#------------------------------------------------------------------------------
# fBpred hullstouch

our $hullstouch_segs = MyTerdragonConf->new
  (name => 'hulls touching',
   segs => [
            "0,0,0",      # around start
            "0,0,1",
            "0,0,2",
            "1,1,2",
            "1,-1,1",
            "-2,0,0",

            "2,0,0",      # around end
            "2,0,1",
            "2,0,2",
            "3,1,2",
            "3,-1,1",
           ]);

our @hullstouch_state_to_conf;
our @hullstouch_state_table;
our ($fNonRpred_hullstouch_flat,$fNonLpred_hullstouch_flat,
     $fNonBpred_hullstouch_flat,
     $fRpred_hullstouch_flat,$fLpred_hullstouch_flat,
     $fBpred_hullstouch_flat);

{
  my %hullstouch_conf_to_state;
  sub hullstouch_conf_to_state {
    my ($conf) = @_;
    my $key = $conf->key;
    my $state = $hullstouch_conf_to_state{$key};
    if (! defined $state) {
      push @hullstouch_state_to_conf, $conf;
      $state = $#hullstouch_state_to_conf;
      $hullstouch_conf_to_state{$key} = $state;
      # print "new state $state\n";
    }
    return $state;
  }
}

our @hullstouch_L_triangles
  = (MyTerdragonConf->new(name => 'hullstouch L top middle',
                       segs => [ '0,0,0', '2,0,1', '1,1,2' ]),
     MyTerdragonConf->new(name => 'hullstouch L top left',
                       segs => [ '-2,0,0', '0,0,1', '-1,1,2' ]),
     MyTerdragonConf->new(name => 'hullstouch L top right',
                       segs => [ '2,0,0', '4,0,1', '3,1,2' ]),
     MyTerdragonConf->new(name => 'hullstouch L bottom left',
                       segs => [ '-1,-1,0', '1,-1,1', '0,0,2' ]),
     MyTerdragonConf->new(name => 'hullstouch L bottom right',
                       segs => [ '1,-1,0', '3,-1,1', '2,0,2' ]),
    );

our @hullstouch_R_triangles
  = (MyTerdragonConf->new(name => 'hullstouch R top left',
                       segs => [ '-1,1,0', '0,0,1', '1,1,2' ]),
     MyTerdragonConf->new(name => 'hullstouch R top right',
                       segs => [ '1,1,0', '2,0,1', '3,1,2' ]),
     MyTerdragonConf->new(name => 'hullstouch R bottom left',
                       segs => [ '-2,0,0', '-1,-1,1', '0,0,2' ]),
     MyTerdragonConf->new(name => 'hullstouch R bottom middle',
                       segs => [ '0,0,0', '1,-1,1', '2,0,2' ]),
     MyTerdragonConf->new(name => 'hullstouch R bottom right',
                       segs => [ '2,0,0', '3,-1,1', '4,0,2' ]),
    );
# foreach my $conf (@hullstouch_L_triangles,@hullstouch_R_triangles){$conf->print;}

# $side==0 for R, $side==1 for L.
# A nearby triangle is boundary of respective $side when it has at least one
# segment $conf but not all of its segments in conf.
sub hullstouch_is_boundary {
  my ($conf, $side) = @_;
  ### hullstouch_is_boundary(): $conf
  $conf->{'within'} // die "oops, no within";
  foreach my $triangle ($side==0 ? @hullstouch_R_triangles
                      : $side==1 ? @hullstouch_L_triangles
                      : die) {
    ### triangle: $triangle->print
    my $got = $triangle->intersection ($conf);
    if ($got->num_segs == 0) {
      # no segments at all around this square, not boundary
      next;
    }
    my $all = $triangle->intersection ($conf->{'within'});
    if ($got->equals($all)) {
      # all segments (of "within") around this square, not boundary
      next;
    }
    # at least 1 but fewer than all segments, so boundary
    return 1;
  }
  return 0;
}

use constant::defer fBpred_hullstouch_init => sub {
  my %options = @_;
  my $verbose = $options{'verbose'} || 0;
  my $verbose_print = $options{'verbose_print'} || sub { print @_; };

  require MyFLAT;
  my @state_enclosed_lists;
  my @hullstouch_state_is_RLB_enclosed;

  hullstouch_conf_to_state(MyTerdragonConf->initial
                           (within => $hullstouch_segs));
  {
    my $state = 0;
    my $depth = 0;
    for (;;) {
      $depth++;
      my $last_state = $#hullstouch_state_to_conf;
      for ( ; $state <= $last_state; $state++) {
        my $from_conf = $hullstouch_state_to_conf[$state];
        if ($verbose >= 3) {
          $verbose_print->("########################################\n");
          $verbose_print->("from $state  ",$from_conf->key,"\n");
          $verbose_print->($from_conf->print_to_str);
        }
        foreach my $digit (0,1,2) {
          my $to_conf = $from_conf->expand($digit);
          my $to_state = hullstouch_conf_to_state($to_conf);
          $hullstouch_state_table[$state]->[$digit] = $to_state;

          if ($verbose >= 3) {
            $verbose_print->("  digit=$digit to $hullstouch_state_table[$state]->[$digit]  ",$to_conf->key,"\n");
            $verbose_print->($to_conf->print_to_str);
          }
        }
      }
      my $new = $#hullstouch_state_to_conf - $last_state;
      if ($verbose) {
        $verbose_print->("depth $depth new $new\n");
      }
      last if $new == 0;
    }
  }

  foreach my $i (0 .. $#hullstouch_state_to_conf) {
    my $conf = $hullstouch_state_to_conf[$i];
    foreach my $side (0, 1) {
      $hullstouch_state_is_RLB_enclosed[$i]->[$side]
        = ! hullstouch_is_boundary($conf,$side);
    }
    $hullstouch_state_is_RLB_enclosed[$i]->[2]
      = $hullstouch_state_is_RLB_enclosed[$i]->[0]
      && $hullstouch_state_is_RLB_enclosed[$i]->[1];
  }
  foreach my $side (0, 1, 2) {
    $state_enclosed_lists[$side]
      = [ grep {$hullstouch_state_is_RLB_enclosed[$_]->[$side]} 0 .. $#hullstouch_state_table ];
  }
  if ($verbose) {
    $verbose_print->("hullstouch enclosed counts ",
                     scalar @{$state_enclosed_lists[0]}, " ",
                     scalar @{$state_enclosed_lists[1]}, " ",
                     scalar @{$state_enclosed_lists[2]});
  }

  if ($verbose) {
    $verbose_print->("  enclosed R = ",join(',',@{$state_enclosed_lists[0]}),
                     " count ",scalar(@{$state_enclosed_lists[0]}),"\n");
    $verbose_print->("  enclosed L = ",join(',',@{$state_enclosed_lists[1]}),
                     " count ",scalar(@{$state_enclosed_lists[1]}),"\n");
    $verbose_print->("  enclosed B = ",join(',',@{$state_enclosed_lists[2]}),
                     " count ",scalar(@{$state_enclosed_lists[2]}),"\n");
  }

  foreach my $side (2, 1, 0) {
    foreach my $i (@{$state_enclosed_lists[$side]}) {
      foreach my $digit (0,1,2) {
        my $to = $hullstouch_state_table[$i]->[$digit];
        unless ($hullstouch_state_is_RLB_enclosed[$to]->[$side]) {
          $hullstouch_state_to_conf[$i]->print;
          $hullstouch_state_to_conf[$i]->expand($digit)->print;
          print "oops, side=$side enclosed $i digit=$digit didn't remain so";
          # print $conf->is_boundary($side);
          exit 1;
        }
      }
    }
  }

  {
    # R
    my $f = MyFLAT::aref_to_FLAT_DFA(\@hullstouch_state_table,
                                     name => 'fNonRpred_hullstouch',
                                     accepting_list => $state_enclosed_lists[0],
                                    );
    # MyFLAT::view($f);
    $f = MyFLAT::eventually_accepting($f, verbose=>$verbose);
    if ($verbose) {
      $verbose_print->("R eventually accepting ",scalar($f->get_accepting),"\n");
    }
    # $f = $f->MyFLAT::minimize;
    $fNonRpred_hullstouch_flat = $f;
    if ($verbose >= 2) {
      $verbose_print->("$f->{name} accepting ",join(' ',$f->get_accepting),"\n");
    }

    $f = $f->complement->MyFLAT::set_name('fRpred_hullstouch');
    $fRpred_hullstouch_flat = $f;
  }
  {
    # L
    my $f = MyFLAT::aref_to_FLAT_DFA(\@hullstouch_state_table,
                                     name => 'fNonLpred_hullstouch',
                                     starting => 0,
                                     accepting_list => $state_enclosed_lists[1],
                                    );
    $f = MyFLAT::eventually_accepting($f, verbose=>$verbose);
    if ($verbose) {
      $verbose_print->("L eventually accepting ",scalar($f->get_accepting),"\n");
    }
    # $f = $f->MyFLAT::minimize;
    $fNonLpred_hullstouch_flat = $f;

    $f = $f->complement->MyFLAT::set_name('fLpred_hullstouch');
    $fLpred_hullstouch_flat = $f;
  }
  {
    # B
    my $f = MyFLAT::aref_to_FLAT_DFA(\@hullstouch_state_table,
                                     name => 'fNonBpred_hullstouch',
                                     starting => 0,
                                     accepting_list => $state_enclosed_lists[2],
                                    );
    $f = MyFLAT::eventually_accepting($f, verbose=>$verbose);
    if ($verbose) {
      $verbose_print->("B eventually accepting ",scalar($f->get_accepting),"\n");
    }
    # $f = $f->MyFLAT::minimize;
    $fNonBpred_hullstouch_flat = $f;
    if ($verbose >= 2) {
      $verbose_print->("$f->{name} accepting ",join(' ',$f->get_accepting),"\n");
    }
    $f = $f->complement->MyFLAT::set_name('fBpred_hullstouch');
    $fBpred_hullstouch_flat = $f;
  }
};


#------------------------------------------------------------------------------
# fBpred hulls around

our $hullsaround_segs_outer = MyTerdragonConf->new
  (name => 'hull outer',
   segs => [
            "-1,-1,1",
            "-1,-1,0",
            "1,-1,0",
            "4,0,2",
            "4,0,1",

            "1,1,0",
            "-1,1,0",
            "-1,1,2",
           ]);

our $hullsaround_segs_inner = MyTerdragonConf->new
  (name => 'hull inner',
   segs => [
            "0,0,0",      # around start
            "0,0,1",
            "0,0,2",
            "1,1,2",
            "1,-1,1",
            "-2,0,0",

            "2,0,0",      # around end
            "2,0,1",
            "2,0,2",
            "3,1,2",
            "3,-1,1",
           ]);

our $hullsaround_segs_all
  = $hullsaround_segs_outer->union($hullsaround_segs_inner)
  ->set_name('hull all');

#------------------------------------------------------------------------------
# the surrounding convex hulls form of fRpred etc

our @hullsaround_state_to_conf;
our @hullsaround_state_table;
our ($fNonRpred_hullsaround_flat,$fNonLpred_hullsaround_flat,
     $fNonBpred_hullsaround_flat,
     $fRpred_hullsaround_flat,$fLpred_hullsaround_flat,
     $fBpred_hullsaround_flat);

{
  my %hullsaround_conf_to_state;
  sub hullsaround_conf_to_state {
    my ($conf) = @_;
    my $key = $conf->key;
    my $state = $hullsaround_conf_to_state{$key};
    if (! defined $state) {
      push @hullsaround_state_to_conf, $conf;
      $state = $#hullsaround_state_to_conf;
      $hullsaround_conf_to_state{$key} = $state;
      # print "new state $state\n";
    }
    return $state;
  }
}

use constant::defer fBpred_hullsaround_init => sub {
  my %options = @_;
  my $verbose = $options{'verbose'} || 0;

  require MyFLAT;
  my @state_enclosed_lists;
  my @hullsaround_state_is_RLB_enclosed;

  hullsaround_conf_to_state(MyTerdragonConf->initial
                            (within => $hullsaround_segs_all));
  {
    my $state = 0;
    my $depth = 0;
    for (;;) {
      $depth++;
      my $last_state = $#hullsaround_state_to_conf;
      for ( ; $state <= $last_state; $state++) {
        my $from_conf = $hullsaround_state_to_conf[$state];
        if ($verbose >= 3) {
          print "########################################\n";
          print "from $state  ",$from_conf->key,"\n";
          $from_conf->print;
        }
        foreach my $digit (0,1,2) {
          my $to_conf = $from_conf->expand($digit);
          my $to_state = hullsaround_conf_to_state($to_conf);
          $hullsaround_state_table[$state]->[$digit] = $to_state;

          if ($verbose >= 3) {
            print "  digit=$digit to $hullsaround_state_table[$state]->[$digit]  ",$to_conf->key,"\n";
            $to_conf->print;
          }
        }
      }
      my $new = $#hullsaround_state_to_conf - $last_state;
      if ($verbose) {
        print "depth $depth new $new\n";
      }
      last if $new == 0;
    }
  }

  foreach my $i (0 .. $#hullsaround_state_to_conf) {
    my $conf = $hullsaround_state_to_conf[$i];
    foreach my $side (0, 1) {
      $hullsaround_state_is_RLB_enclosed[$i]->[$side]
        = ! $conf->is_boundary($side);
    }
    $hullsaround_state_is_RLB_enclosed[$i]->[2]
      = $hullsaround_state_is_RLB_enclosed[$i]->[0]
      && $hullsaround_state_is_RLB_enclosed[$i]->[1];
  }
  foreach my $side (0, 1, 2) {
    $state_enclosed_lists[$side]
      = [ grep {$hullsaround_state_is_RLB_enclosed[$_]->[$side]} 0 .. $#hullsaround_state_table ];
  }
  if ($verbose) {
    print "hullsaround enclosed counts ",
      scalar @{$state_enclosed_lists[0]}, " ",
      scalar @{$state_enclosed_lists[1]}, " ",
      scalar @{$state_enclosed_lists[2]};
  }

  if ($verbose) {
    print "  enclosed R = ",join(',',@{$state_enclosed_lists[0]}),
      " count ",scalar(@{$state_enclosed_lists[0]}),"\n";
    print "  enclosed L = ",join(',',@{$state_enclosed_lists[1]}),
      " count ",scalar(@{$state_enclosed_lists[1]}),"\n";
    print "  enclosed B = ",join(',',@{$state_enclosed_lists[2]}),
      " count ",scalar(@{$state_enclosed_lists[2]}),"\n";
  }

  foreach my $side (2, 1, 0) {
    foreach my $i (@{$state_enclosed_lists[$side]}) {
      foreach my $digit (0,1,2) {
        my $to = $hullsaround_state_table[$i]->[$digit];
        unless ($hullsaround_state_is_RLB_enclosed[$to]->[$side]) {
          $hullsaround_state_to_conf[$i]->print;
          $hullsaround_state_to_conf[$i]->expand($digit)->print;
          print "oops, side=$side enclosed $i digit=$digit didn't remain so";
          # print $conf->is_boundary($side);
          exit 1;
        }
      }
    }
  }

  {
    # R
    my $f = MyFLAT::aref_to_FLAT_DFA(\@hullsaround_state_table,
                                     name => 'fNonRpred_hullsaround',
                                     accepting_list => $state_enclosed_lists[0],
                                    );
    # MyFLAT::view($f);
    $f = MyFLAT::eventually_accepting($f, verbose=>$verbose);
    if ($verbose) {
      print "R eventually accepting ",scalar($f->get_accepting),"\n";
    }
    # $f = $f->MyFLAT::minimize;
    $fNonRpred_hullsaround_flat = $f;
    if ($verbose >= 2) {
      print "$f->{name} accepting ",join(' ',$f->get_accepting),"\n";
    }

    $f = $f->complement->MyFLAT::set_name('fRpred_hullsaround');
    $fRpred_hullsaround_flat = $f;
  }
  {
    # L
    my $f = MyFLAT::aref_to_FLAT_DFA(\@hullsaround_state_table,
                                     name => 'fNonLpred_hullsaround',
                                     starting => 0,
                                     accepting_list => $state_enclosed_lists[1],
                                    );
    $f = MyFLAT::eventually_accepting($f, verbose=>$verbose);
    if ($verbose) {
      print "L eventually accepting ",scalar($f->get_accepting),"\n";
    }
    # $f = $f->MyFLAT::minimize;
    $fNonLpred_hullsaround_flat = $f;

    $f = $f->complement->MyFLAT::set_name('fLpred_hullsaround');
    $fLpred_hullsaround_flat = $f;
  }
  {
    # B
    my $f = MyFLAT::aref_to_FLAT_DFA(\@hullsaround_state_table,
                                     name => 'fNonBpred_hullsaround',
                                     starting => 0,
                                     accepting_list => $state_enclosed_lists[2],
                                    );
    $f = MyFLAT::eventually_accepting($f, verbose=>$verbose);
    if ($verbose) {
      print "B eventually accepting ",scalar($f->get_accepting),"\n";
    }
    # $f = $f->MyFLAT::minimize;
    $fNonBpred_hullsaround_flat = $f;
    if ($verbose >= 2) {
      print "$f->{name} accepting ",join(' ',$f->get_accepting),"\n";
    }
    $f = $f->complement->MyFLAT::set_name('fBpred_hullsaround');
    $fBpred_hullsaround_flat = $f;
  }

};

#------------------------------------------------------------------------------
# fRpred etc as explicit state machines
# (the same as hullsaround but much smaller)

our @fRpred_state_table
  = ([0,2,1], [3,2,1], [0,4,5], [5,5,3], [4,5,5], [5,5,5]);

# fNonRpred_flat() matches initial strings of digits of a fractional f which
# is not on the right boundary (no matter what follows).
#
# fRpred_flat() is the complement.  It matches initial strings of digits of
# a fractional f which can be on the right boundary, ie. there exists
# further digits below which are an f on the right boundary.
#
use constant::defer fNonRpred_flat => sub {
  require MyFLAT;
  MyFLAT::aref_to_FLAT_DFA(\@fRpred_state_table, name => 'fNonRpred');
};
use constant::defer fRpred_flat => sub {
  require MyFLAT;
  return fNonRpred_flat()->complement->MyFLAT::set_name('fRpred');
};

our @fLpred_state_table
  = ([1,2,0], [1,2,3], [5,4,0], [3,5,5], [5,5,4], [5,5,5]);

use constant::defer fNonLpred_flat => sub {
  require MyFLAT;
  MyFLAT::aref_to_FLAT_DFA(\@fLpred_state_table, name => 'fNonLpred');
};
use constant::defer fLpred_flat => sub {
  require MyFLAT;
  return fNonLpred_flat()->complement->MyFLAT::set_name('fLpred');
};

use constant::defer fNonBpred_flat => sub {
  require MyFLAT;
  return fNonRpred_flat()->intersect(fNonLpred_flat())->as_dfa
    ->MyFLAT::set_name('fNonBpred');
};
use constant::defer fBpred_flat => sub {
  require MyFLAT;
  return fRpred_flat()->union(fLpred_flat())->as_dfa
    ->MyFLAT::set_name('fBpred');
};

#------------------------------------------------------------------------------
# Graph.pm

# return a Graph.pm graph of terdragon curve level $k
sub terdragon_Graph {
  my ($k, %options) = @_;
  my $path = TerdragonCurve_path();
  require Graph;
  my $graph = Graph->new (undirected => 1);
  $graph->set_graph_attribute (name => "Terdragon $k");
  $graph->set_graph_attribute('vertex_name_type','xy-triangular');
  my ($n_lo, $n_hi) = $path->level_to_n_range($k);
  if (defined $options{'n_lo'}) { $n_lo = $options{'n_lo'}; }
  if (defined $options{'n_hi'}) { $n_hi = $options{'n_hi'}; }
  my ($prev_x,$prev_y) = $path->n_to_xy($n_lo);
  $graph->add_vertex("$prev_x,$prev_y");
  foreach my $n ($n_lo+1 .. $n_hi) {
    my ($x,$y) = $path->n_to_xy($n);
    $graph->add_edge("$prev_x,$prev_y", "$x,$y");
    ($prev_x,$prev_y) = ($x,$y);
  }
  return $graph;
}

# Return a Graph.pm graph of the right turns in a triangle of terdragons.
sub triangle_turn_tree_Graph {
  my ($k) = @_;
  require Graph;
  my $path = TerdragonCurve_path();
  my $graph = Graph->new(undirected => 1);
  $graph->set_graph_attribute (name => "Terdragon Triangle Turn Tree $k");
  $graph->set_graph_attribute('vertex_name_type','xy-triangular');
  my $n_lo = 2*3**$k;
  my $n_hi = 5*3**$k;
  my ($sx,$sy) = $path->n_to_xy($n_lo);
  my $prev;
  foreach my $n ($n_lo .. $n_hi-1) {
    my $turn = turn($n);
    if ($n==$n_lo) { $turn = 1; }
    if ($turn == -1) {
      my ($x,$y) = $path->n_to_xy($n);
      $x -= $sx;
      $y -= $sy;
      my $v = "$x,$y";
      $graph->add_vertex($v);
      if (defined $prev) { $graph->add_edge($v,$prev); }
      $prev = $v;
    } else {
      undef $prev;
    }
  }
  return $graph;
}

# Return a Graph.pm graph for unit triangles and their connections for a
# triangle of terdragons.
# Vertices are unit triangles on the inside of the curve, and also points
# (corners of the triangles) where those triangles are consecutive in the
# curve.
# Edges are between a triangle and each of its corner connection points.
#
sub triangle_area_and_connections_tree_Graph {
  my ($k) = @_;
  require Graph;
  my $graph = Graph->new(undirected => 1);
  $graph->set_graph_attribute (name => "Terdragon Area and Connections Tree $k");
  $graph->set_graph_attribute('vertex_name_type','xy-triangular');

  my $path = TerdragonCurve_path();
  my $n_lo = 2*3**$k;
  my $n_hi = 5*3**$k;
  my ($sx,$sy) = $path->n_to_xy($n_lo);
  foreach my $n ($n_lo .. $n_hi-1) {
    my ($x,$y) = $path->n_to_xy($n);
    my ($x2,$y2) = $path->n_to_xy($n+1);
    my ($dx,$dy) = $path->n_to_dxdy($n);
    $x -= $sx;
    $y -= $sy;
    $x2 -= $sx;
    $y2 -= $sy;
    ($x,$y) = mul($x,$y);
    ($x2,$y2) = mul($x2,$y2);
    my $v = "$x,$y";
    my $v2 = "$x2,$y2";
    $x2 -= $dx;
    $y2 -= $dy;
    my $vm = "$x2,$y2";

    my $t  = ($n!=$n_lo   && turn($n)   == -1);
    my $t2 = ($n+1!=$n_hi && turn($n+1) == -1);
    if ($t)  { $graph->add_edge($v,$vm); }
    if ($t2) { $graph->add_edge($vm,$v2); }
    $graph->add_vertex($vm);
    ### $n
    ### $t
    ### $t2
    # $graph->add_path($v,$vm,$v2);
  }
  return $graph;
}

# Return a Graph.pm graph for unit triangles inside a triangle of terdragons.
# Each vertex is a unit triangle.
# Edges are between those which are consecutive in the curve.
sub triangle_area_Graph {
  my ($k) = @_;
  require Graph;
  my $graph = Graph->new(undirected => 1);
  $graph->set_graph_attribute (name => "Terdragon Area $k");
  $graph->set_graph_attribute('vertex_name_type','xy-triangular');

  my $path = TerdragonCurve_path();
  my $n_lo = 2*3**$k;
  my $n_hi = 5*3**$k;
  my ($sx,$sy) = $path->n_to_xy($n_lo);
  my $prev;
  foreach my $n ($n_lo .. $n_hi-1) {
    my ($x,$y) = $path->n_to_xy($n);
    my ($dx,$dy) = $path->n_to_dxdy($n);
    $x -= $sx;
    $y -= $sy;
    ($dx,$dy) = xy_rotate_plus60($dx,$dy);
    ($x,$y) = mul($x,$y);
    $x += $dx;
    $y += $dy;
    my $v = "$x,$y";
    $graph->add_vertex($v);
    if (defined $prev && $v ne $prev) { $graph->add_edge($v,$prev); }
    $prev = $v;
  }
  return $graph;

  # my xydxdy_to_right_triangle_name = sub {
  #   my ($x,$y, $dx,$dy) = @_;
  #   ($x,$y) = mul($x,$y);
  #   $x += $dx;
  #   $y += $dy;
  #   return "$x,$y";
  # };
}


#------------------------------------------------------------------------------


our @AltRpred9_state_table
  = ([0,4,4,4,1,2,0,1,2],  # 0 = R
     [0,4,4,4,1,2,4,4,4],
     [4,4,4,4,4,4,0,1,2],
     [0,1,2,4,4,4,4,4,4],  # 3 = O
     [4,4,4,4,4,4,4,4,4]);  # 4 = non

use constant::defer AltNonRpred9_flat_even => sub {
  require MyFLAT;
  MyFLAT::aref_to_FLAT_DFA(\@AltRpred9_state_table, name => 'AltNonRpred9');
};
use constant::defer AltRpred9_flat_even => sub {
  require MyFLAT;
  return AltNonRpred9_flat_even()->complement->MyFLAT::set_name('AltRpred9');
};


our @AltRpred3_state_table
  = ([5,4,3],  # 0
     [5,4,6],
     [6,6,3],
     [0,1,2],  # 3 even
     [6,1,2],  # 4
     [0,6,6],  # 5
     [6,6,6]);

use constant::defer AltNonRpred3_flat_even => sub {
  require MyFLAT;
  MyFLAT::aref_to_FLAT_DFA(\@AltRpred3_state_table,
                           starting => 0,
                           name => 'AltNonRpred3');
};
use constant::defer AltNonRpred3_flat_odd => sub {
  require MyFLAT;
  MyFLAT::aref_to_FLAT_DFA(\@AltRpred3_state_table,
                           starting => 3,
                           name => 'AltNonRpred3');
};
use constant::defer AltNonRpred3_flat => sub {
  require MyFLAT;
  require FLAT;
  my $even = AltNonRpred3_flat_even()
    ->intersect(FLAT::Regex->new('        ( (0|1|2) (0|1|2) )*')->as_dfa);
  my $odd  = AltNonRpred3_flat_odd()
    ->intersect(FLAT::Regex->new('(0|1|2) ( (0|1|2) (0|1|2) )*')->as_dfa);
  $even->union($odd)->MyFLAT::minimize
    ->MyFLAT::set_name('AltNonRpred3');
};
use constant::defer AltRpred3_flat => sub {
  require MyFLAT;
  return AltNonRpred3_flat()->complement->MyFLAT::set_name('AltRpred3');
};

sub AltRpred_k_of_ternary_str {
  my ($str) = @_;
  return AltRpred3_flat->contains($str);
}
sub AltRpred_k {
  my ($k,$n) = @_;
  $n >= 0 || croak 'AltRpred_k() is for n>=0';
  $n = cnv_padded($n,10,3,$k);
  return (length($n) == $k && AltRpred_k_of_ternary_str($n));
}
sub AltLpred_k {
  my ($k,$n) = @_;
  return AltRpred_k($k, 3**$k-1-$n);
}


#------------------------------------------------------------------------------
# fBpred hullstouch

our $alt_hullstouch_segs = MyTerdragonConf->new
  (name => 'alt hulls touching',
   segs => [
            '0,0,0',      # around start
            '0,0,1',
            '0,0,2',
            '1,1,2',
            '1,-1,1',
            '-2,0,0',

            '2,0,0',      # around end
            '2,0,1',
            '2,0,2',
            '3,1,2',
            '3,-1,1',

            '-1,1,0',     # across
            '1,-1,0',
           ]);

our $alt_hullstouch_odd_segs = MyTerdragonConf->new
  (name => 'alt hulls touching',
   segs => [
            '0,0,0',      # around start
            '0,0,1',
            '0,0,2',
            '1,1,2',
            '1,-1,1',
            '-2,0,0',

            '2,0,0',      # around end
            '2,0,1',
            '2,0,2',
            '3,1,2',
            '3,-1,1',

            '-1,-1,0',     # across
            '1,1,0',
           ]);

our @alt_hullstouch_state_to_conf;
our @alt_hullstouch_state_table;
our ($fAltNonRpred_hullstouch_flat,
     $fAltNonLpred_hullstouch_flat,
     $fAltNonBpred_hullstouch_flat,
     $fAltRpred_hullstouch_flat,
     $fAltLpred_hullstouch_flat,
     $fAltBpred_hullstouch_flat);

{
  my %alt_hullstouch_conf_to_state;
  sub alt_hullstouch_conf_to_state {
    my ($conf) = @_;
    my $key = $conf->key;
    my $state = $alt_hullstouch_conf_to_state{$key};
    if (! defined $state) {
      push @alt_hullstouch_state_to_conf, $conf;
      $state = $#alt_hullstouch_state_to_conf;
      $conf->set_name("state $state");
      $alt_hullstouch_conf_to_state{$key} = $state;
      # print "new state $state\n";
    }
    return $state;
  }
}

use constant::defer fAltBpred_hullstouch_init => sub {
  my %options = @_;
  my $verbose = $options{'verbose'} || 0;
  my $verbose_print = $options{'verbose_print'} || sub { print @_; };

  require MyFLAT;
  my @state_enclosed_lists;
  my @alt_hullstouch_state_is_RLB_enclosed;

  alt_hullstouch_conf_to_state(MyTerdragonConf->initial
                               (within => $alt_hullstouch_segs,
                                odd => 0));
  {
    my $state = 0;
    my $depth = 0;
    for (;;) {
      $depth++;
      my $last_state = $#alt_hullstouch_state_to_conf;
      for ( ; $state <= $last_state; $state++) {
        my $from_conf = $alt_hullstouch_state_to_conf[$state];
        if ($verbose >= 3) {
          $verbose_print->("########################################\n");
          $verbose_print->("from $state  ",$from_conf->key,"\n");
          $verbose_print->($from_conf->print_to_str);
        }
        foreach my $digit (0 .. 8) {
          my $to_conf = $from_conf->alt_expand9($digit);
          my $to_state = alt_hullstouch_conf_to_state($to_conf);
          $alt_hullstouch_state_table[$state]->[$digit] = $to_state;

          if ($verbose >= 3) {
            $verbose_print->("  digit=$digit to $alt_hullstouch_state_table[$state]->[$digit]  ",$to_conf->key,"\n");
            $verbose_print->($to_conf->print_to_str);
          }
        }
      }
      my $new = $#alt_hullstouch_state_to_conf - $last_state;
      if ($verbose) {
        $verbose_print->("depth $depth new $new\n");
      }
      last if $new == 0;
    }
  }

  foreach my $i (0 .. $#alt_hullstouch_state_to_conf) {
    my $conf = $alt_hullstouch_state_to_conf[$i];
    foreach my $side (0, 1) {
      $alt_hullstouch_state_is_RLB_enclosed[$i]->[$side]
        = ! hullstouch_is_boundary($conf,$side);
    }
    $alt_hullstouch_state_is_RLB_enclosed[$i]->[2]
      = $alt_hullstouch_state_is_RLB_enclosed[$i]->[0]
      && $alt_hullstouch_state_is_RLB_enclosed[$i]->[1];
  }
  foreach my $side (0, 1, 2) {
    $state_enclosed_lists[$side]
      = [ grep {$alt_hullstouch_state_is_RLB_enclosed[$_]->[$side]}
          0 .. $#alt_hullstouch_state_table ];
  }
  if ($verbose) {
    $verbose_print->("alt_hullstouch enclosed counts ",
                     scalar @{$state_enclosed_lists[0]}, " ",
                     scalar @{$state_enclosed_lists[1]}, " ",
                     scalar @{$state_enclosed_lists[2]});
  }

  if ($verbose) {
    $verbose_print->("  enclosed R = ",join(',',@{$state_enclosed_lists[0]}),
                     " count ",scalar(@{$state_enclosed_lists[0]}),"\n");
    $verbose_print->("  enclosed L = ",join(',',@{$state_enclosed_lists[1]}),
                     " count ",scalar(@{$state_enclosed_lists[1]}),"\n");
    $verbose_print->("  enclosed B = ",join(',',@{$state_enclosed_lists[2]}),
                     " count ",scalar(@{$state_enclosed_lists[2]}),"\n");
  }

  foreach my $side (2, 1, 0) {
    foreach my $i (@{$state_enclosed_lists[$side]}) {
      foreach my $digit (0 .. 8) {
        my $to = $alt_hullstouch_state_table[$i]->[$digit];
        unless ($alt_hullstouch_state_is_RLB_enclosed[$to]->[$side]) {
          print "oops, side=$side enclosed $i digit=$digit didn't remain so\n";
          print "From:\n";
          $alt_hullstouch_state_to_conf[$i]->print;
          print "To:\n";
          $alt_hullstouch_state_to_conf[$i]->alt_expand9($digit)->print;
          # print $conf->is_boundary($side);
          # exit 1;
        }
      }
    }
  }

  {
    # R
    my $f = MyFLAT::aref_to_FLAT_DFA(\@alt_hullstouch_state_table,
                                     name => 'fAltNonRpred_hullstouch',
                                     accepting_list => $state_enclosed_lists[0],
                                    );
    # MyFLAT::view($f);
    $f = MyFLAT::eventually_accepting($f, verbose=>$verbose);
    if ($verbose) {
      $verbose_print->("R eventually accepting ",scalar($f->get_accepting),"\n");
    }
    # $f = $f->MyFLAT::minimize;
    $fAltNonRpred_hullstouch_flat = $f;
    if ($verbose >= 2) {
      $verbose_print->("$f->{name} accepting ",join(' ',$f->get_accepting),"\n");
    }

    $f = $f->complement->MyFLAT::set_name('fAltRpred_hullstouch');
    $fAltRpred_hullstouch_flat = $f;
  }
  {
    # L
    my $f = MyFLAT::aref_to_FLAT_DFA(\@alt_hullstouch_state_table,
                                     name => 'fAltNonLpred_hullstouch',
                                     starting => 0,
                                     accepting_list => $state_enclosed_lists[1],
                                    );
    $f = MyFLAT::eventually_accepting($f, verbose=>$verbose);
    if ($verbose) {
      $verbose_print->("L eventually accepting ",scalar($f->get_accepting),"\n");
    }
    # $f = $f->MyFLAT::minimize;
    $fAltNonLpred_hullstouch_flat = $f;

    $f = $f->complement->MyFLAT::set_name('fAltLpred_hullstouch');
    $fAltLpred_hullstouch_flat = $f;
  }
  {
    # B
    my $f = MyFLAT::aref_to_FLAT_DFA(\@alt_hullstouch_state_table,
                                     name => 'fAltNonBpred_hullstouch',
                                     starting => 0,
                                     accepting_list => $state_enclosed_lists[2],
                                    );
    $f = MyFLAT::eventually_accepting($f, verbose=>$verbose);
    if ($verbose) {
      $verbose_print->("B eventually accepting ",scalar($f->get_accepting),"\n");
    }
    # $f = $f->MyFLAT::minimize;
    $fAltNonBpred_hullstouch_flat = $f;
    if ($verbose >= 2) {
      $verbose_print->("$f->{name} accepting ",join(' ',$f->get_accepting),"\n");
    }
    $f = $f->complement->MyFLAT::set_name('fAltBpred_hullstouch');
    $fAltBpred_hullstouch_flat = $f;
  }
};

our @fAltBpred_state_table
  = ([3,2,1], [0,5,4], [8,7,6], [0,10,9], [9,11,1], [14,13,12], [15,11,1],
     [3,13,12], [9,9,8], [9,9,9], [16,9,9], [9,9,17], [12,9,9], [8,7,4],
     [0,10,18], [9,9,18], [12,9,18], [15,9,8], [15,9,9]);
our @fAltBpred_state_names
  = ('Re',      # 0
     'Ro',      # 1
     'Mo-dash', # 2
     'Eo',      # 3
     'Ee',      # 4
     'Me-dash', # 5
     'Ee-dash', # 6
     'Me',      # 7
     'T',       # 8
     'non',     # 9
     'Eo1',     # 10
     'Ee1',     # 11
     'Z',       # 12
     'Mo',      # 13
     'Eo-dash', # 14
     'Q0',      # 15
     'Eo2',     # 16
     'Ee2',     # 17
     'Q2',      # 18
    );
our @fAltRpred_accepting
  = grep {$fAltBpred_state_names[$_] ne 'non'} 0 .. $#fAltBpred_state_table;

use constant::defer fAltRpred_flat => sub {
  require MyFLAT;
  MyFLAT::aref_to_FLAT_DFA(\@fAltBpred_state_table,
                           starting => 0,
                           accepting_list => \@fAltRpred_accepting,
                           name => 'fAltRpred');
};

#------------------------------------------------------------------------------

# Local variables:
# compile-command: "perl -w MyTerdragon.pm"
# End:

1;
__END__
